#include <gmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <omp.h>

struct MPZ_POOL
{
    int head;
    int count;
    int length;
    mpz_t *array;
    omp_lock_t *lock;
};

typedef struct MPZ_POOL mpz_pool_t[1];


void mpz_pool_init(mpz_pool_t pool, int initialSize);
void mpz_pool_clear(mpz_pool_t pool);
void mpz_pool_get(mpz_pool_t pool, mpz_t val);
void mpz_pool_get_array(mpz_pool_t pool, mpz_t * val, int n);
void mpz_pool_release(mpz_pool_t pool, mpz_t val);
void mpz_pool_release_array(mpz_pool_t pool, mpz_t * val, int n);
void mpz_pool_realloc(mpz_pool_t pool);


void mpz_pool_init(mpz_pool_t pool, int initialSize)
{
    int i;
    pool->array = (mpz_t * )malloc(sizeof(mpz_t) * initialSize);
    pool->head = 0;
    pool->count = initialSize;
    pool->length = initialSize;
    pool->lock = (omp_lock_t * )malloc(sizeof(omp_lock_t));
    omp_init_lock(pool->lock);
    
    for(i = 0;i < initialSize;i++)
    {
        mpz_init((pool->array)[i]);
    }
}

void mpz_pool_clear(mpz_pool_t pool)
{
    int i;
    if(pool->length > pool->count)
    {
        printf("%d objects are not released!!!", pool->length - pool->count);
    }
    
    for(i = 0;i < pool->length;i++)
    {
        mpz_clear((pool->array)[i]);
    }
    
    free(pool->array);
    free(pool->lock);
}

void mpz_pool_get(mpz_pool_t pool, mpz_t val)
{
    omp_set_lock(pool->lock);
    if(pool->count == 0)
    {
        pool->head = pool->length;
        mpz_pool_realloc(pool);
    }

    if(pool->head >= pool->length)
    {
        pool->head = 0;
    }
    
    val[0] = (pool->array)[pool->head][0];
    pool->head++;
    pool->count--;
    omp_unset_lock(pool->lock);
}

void mpz_pool_get_array(mpz_pool_t pool, mpz_t * val, int n)
{
    int i;
    for(i = 0;i < n;i++)
    {
        mpz_pool_get(pool, val[i]);
    }
    return;
    // there is bug. please fix!!
    /*if(pool->count >= n)
    {
        int copyCount = pool->length - pool->head;
        if(copyCount >= n)
        {
            memcpy(val, pool->array + pool->head, sizeof(mpz_t) * n);
            pool->head += n;
            pool->count -= n;
            return;
        }
        memcpy(val, pool->array + pool->head, sizeof(mpz_t) * copyCount);
        val += copyCount;
        pool->head = 0;
        pool->count -= copyCount;
        copyCount = n - copyCount;
        memcpy(val, pool->array, sizeof(mpz_t) * copyCount);
        pool->head += copyCount;
        pool->count -= copyCount;
        return;
    }
    
    int copyCount = pool->length - pool->head;
    memcpy(val, pool->array + pool->head, sizeof(mpz_t) * copyCount);
    pool->head += copyCount;
    pool->count -= copyCount;
    val += copyCount;
    n -= copyCount;
    if(pool->count > 0)
    {
        copyCount = pool->count;
        memcpy(val, pool->array, sizeof(mpz_t) * copyCount);
        pool->count -= copyCount;
        pool->head = pool->length;
        val += copyCount;
        n -= copyCount;
    }
    while(n > pool->count)
    {
        mpz_pool_realloc(pool);
    }
    memcpy(val, pool->array + pool->head, sizeof(mpz_t) * n);
    pool->head += n;
    pool->count -= n;*/
}

void mpz_pool_release(mpz_pool_t pool, mpz_t val)
{
    omp_set_lock(pool->lock);
    int tail = pool->head + pool->count;
    if(tail >= pool->length)
    {
        tail -= pool->length;
    }
    (pool->array)[tail][0] = val[0];
    pool->count++;
    omp_unset_lock(pool->lock);
}

void mpz_pool_release_array(mpz_pool_t pool, mpz_t * val, int n)
{
    int i;
    for(i = 0;i < n;i++)
    {
        mpz_pool_release(pool, val[i]);
    }
    return;
    // there is bug. please fix!!
    /*int copyCount, tail = pool->head + pool->count;
    if(tail >= pool->length)
    {
        tail -= pool->length;
    }
    
    copyCount = pool->length - tail;
    if(copyCount >= n)
    {
        memcpy(pool->array + tail, val, sizeof(mpz_t) * n);
        pool->count += n;
    }else{
        memcpy(pool->array + tail, val, sizeof(mpz_t) * copyCount);
        val += copyCount;
        copyCount = n - copyCount;
        memcpy(pool->array, val, sizeof(mpz_t) * copyCount);
        pool->count += n;
    }*/
}

void mpz_pool_realloc(mpz_pool_t pool)
{
    int prevSize, newSize, i;
    prevSize = pool->length;
    newSize = prevSize * 2;
    if(newSize <= 0)
    {
        newSize = 4;
    }
    pool->array = (mpz_t * )realloc(pool->array, sizeof(mpz_t) * newSize);
    pool->length = newSize;
    pool->count += newSize - prevSize;
    for(i = prevSize;i < newSize;i++)
    {
        mpz_init((pool->array)[i]);
    }
}